/* Everything about the rules for NAT. */
#define __NO_VERSION__
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/bitops.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_nat_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_nat_lock)

#include "NAT/ip_nat_rule.h"
#include "NAT/ip_nat_protocol.h"
#include "common/listhelp.h"

#ifdef DEBUG_CONNTRACK
#define static
#endif

static LIST_HEAD(nat_rules);
static LIST_HEAD(mapping_types);

static inline int lookup_cmp(const struct ip_nat_rule *ptr,
			     const struct ip_conntrack_tuple *tuple)
{
	size_t i;
	unsigned long cmp = 0;

	if (ptr->dead)
		return 0;
	/* Apply mask.  Should unroll nicely */
	for (i = 0; i < sizeof(*tuple)/sizeof(unsigned long); i++) {
		cmp |= (((unsigned long *)tuple)[i]
			& ((unsigned long *)&ptr->mask)[i])
			^ ((unsigned long *)&ptr->match)[i];
	}

	return (cmp == 0);
}

struct ip_nat_rule *
__ip_nat_rule_get(const struct ip_conntrack_tuple *tuple)
{
	struct ip_nat_rule *rule;

	MUST_BE_READ_LOCKED(&ip_nat_lock);
	rule = LIST_FIND(&nat_rules, lookup_cmp, struct ip_nat_rule *, tuple);
	if (rule)
		atomic_inc(&rule->use);

	return rule;
}	

struct ip_nat_rule *
ip_nat_rule_get(const struct ip_conntrack_tuple *tuple)
{
	struct ip_nat_rule *rule;

	READ_LOCK(&ip_nat_lock);
	rule = __ip_nat_rule_get(tuple);
	READ_UNLOCK(&ip_nat_lock);

	return rule;
}

void
ip_nat_rule_put(struct ip_nat_rule *rule)
{
	if (atomic_dec_and_test(&rule->use)) {
		IP_CONNTRACK_ASSERT(rule->dead);
		WRITE_LOCK(&ip_nat_lock);
		LIST_DELETE(&nat_rules, rule);
		WRITE_UNLOCK(&ip_nat_lock);
		kfree(rule);
	}
}

/* NAT rule tables are ordered, similar to routing tables: most
 * specific to least specific.  I've tried (and maybe failed) to be
 * intuitive here.  Higher the better.  This is a partial ordering.  */
static u_int32_t
precedence(const struct ip_conntrack_tuple *mask)
{
	u_int32_t prec = 0;

	/* We build this up most important to least. */

	/* Source IP bits + destination IP bits: 7 bits */
	prec |= hweight32(mask->src.ip) + hweight32(mask->dst.ip);

	/* Do we care about protocol?: 1 bit */
	prec <<= 1;
	prec |= mask->dst.protonum ? 1 : 0;

	/* Per-protocol information: 6 bits */
	prec <<= 6;
	prec |= hweight16(mask->src.u.all) + hweight16(mask->dst.u.all);

	return prec;
}

#ifndef ALWAYS_IN_INTERRUPT
/* Must NOT have lock. */
static void load_mapping_type(const char *name)
{
	char modulename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN+sizeof("ip_nat_map_")];

	strcpy(modulename, "ip_nat_map_");
	strcat(modulename, name);
	DEBUGP("ip_nat: requesting module `%s'\n", modulename);
	request_module(modulename);
}
#endif

int ip_nat_register_mapping_type(struct ip_nat_mapping_type *me)
{
	int ret;

	WRITE_LOCK(&ip_nat_lock);
	if (list_named_find(&mapping_types, me->name))
		ret = -EEXIST;
	else {
		MOD_INC_USE_COUNT;
		list_prepend(&mapping_types, me);
		ret = 0;
	}
	WRITE_UNLOCK(&ip_nat_lock);

	return ret;
}

void ip_nat_unregister_mapping_type(struct ip_nat_mapping_type *me)
{
	WRITE_LOCK(&ip_nat_lock);
	LIST_DELETE(&mapping_types, me);
	WRITE_UNLOCK(&ip_nat_lock);

	synchronize_bh();
	MOD_DEC_USE_COUNT;
}

/* We allocate extra stuff for the expect rule. */
struct expect_rule {
	struct ip_nat_rule rule;
	struct expect_rule *moptions;
	struct timer_list timer;
};

/* We create an expectation by adding a rule with a special mapping type;
   rule has a struct timer_list on the end. */
static unsigned int 
expect_setup(struct ip_conntrack *conntrack,
	     enum ip_nat_manip_type maniptype,
	     struct ip_nat_multi_range *mr,
	     const struct net_device *device,
	     void *moptions)
{
	unsigned int ret;
	struct expect_rule *rule = *((struct expect_rule **)moptions);

	/* Beware: timer may have raced us.  Check. */
	if (del_timer(&rule->timer)) {
		WRITE_LOCK(&ip_nat_lock);
		ret = ip_nat_setup_info(conntrack, maniptype, mr, device);

		/* Delete the rule which created us. */
		LIST_DELETE(&nat_rules, rule);
		WRITE_UNLOCK(&ip_nat_lock);
		kfree(rule);
	} else {
		/* FIXME: Should really keep iterating down other
		   rules here, but it's so rare. --RR */
		ret = NF_DROP;
	}
	return ret;
}

static struct ip_nat_mapping_type mapping_expect
= { { NULL, NULL }, "expected", NULL, NULL, expect_setup, 
    sizeof(struct ip_nat_rule *) };

static void timeout_rule(unsigned long ul_rule)
{
	struct expect_rule *rule = (struct expect_rule *)ul_rule;

	DEBUGP("timeout_rule: giving up on expected %p.\n", rule);
	/* Use write lock to protect against race with expect_setup above */
	WRITE_LOCK(&ip_nat_lock);
	if (list_inlist(&nat_rules, &rule->rule)) 
		LIST_DELETE(&nat_rules, &rule->rule);
	else {
		DEBUGP("timeout_rule: lost race for %p.\n", rule);
		rule = NULL;
	}
	WRITE_UNLOCK(&ip_nat_lock);
	if (rule) 
		kfree(rule);
}

/* Returns TRUE if it successfully creates expectation. */
int ip_nat_expect(const struct ip_conntrack_tuple *expect,
		  const struct ip_conntrack_manip *manip,
		  enum ip_nat_manip_type maniptype,
		  unsigned long timeout)
{
	struct expect_rule *expect_rule;

	expect_rule = kmalloc(sizeof(*expect_rule), GFP_ATOMIC);
	if (!expect_rule) {
		DEBUGP("ip_nat_expect: oom!\n");
		return 0;
	}

	expect_rule->rule = ((struct ip_nat_rule) 
			     { { NULL, NULL }, ATOMIC_INIT(1), 0, *expect,
			       { { 0xFFFFFFFF, { 0 }, 0 },
				 { 0xFFFFFFFF, { 0xFFFF }, 0xFF } },
			       maniptype, 0,
			       &mapping_expect, &expect_rule->moptions,
			       { 1, 
				 { { IP_NAT_RANGE_MAP_IPS
				     | IP_NAT_RANGE_PROTO_SPECIFIED,
				     manip->ip, manip->ip,
				     manip->u, manip->u } }
			       } });
	expect_rule->moptions = expect_rule;
	init_timer(&expect_rule->timer);
	expect_rule->timer.data = (unsigned long)expect_rule;
	expect_rule->timer.function = timeout_rule;
	expect_rule->timer.expires = jiffies + timeout;
	add_timer(&expect_rule->timer);

#if 0
	DEBUGP("ip_nat_expect: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u proto %u.\n"
	       "ip_nat_mask:   %u.%u.%u.%u:%u->%u.%u.%u.%u:%u proto %u.\n",
	       IP_PARTS(expect_rule->rule.match.src.ip), 
	       ntohs(expect_rule->rule.match.src.u.all),
	       IP_PARTS(expect_rule->rule.match.dst.ip), 
	       ntohs(expect_rule->rule.match.dst.u.all),
	       expect_rule->rule.match.dst.protonum,
	       IP_PARTS(expect_rule->rule.mask.src.ip), 
	       ntohs(expect_rule->rule.mask.src.u.all),
	       IP_PARTS(expect_rule->rule.mask.dst.ip), 
	       ntohs(expect_rule->rule.mask.dst.u.all),
	       expect_rule->rule.mask.dst.protonum);
#endif

	/* Expect rules must take precendence. */
	WRITE_LOCK(&ip_nat_lock);
	list_prepend(&nat_rules, &expect_rule->rule);
	WRITE_UNLOCK(&ip_nat_lock);
	return 1;
}

static inline int prec_cmpfn(struct ip_nat_rule *new, struct ip_nat_rule *i)
{
	return (new->precedence > i->precedence);
}

static int
rule_usercmpfn(const struct ip_nat_rule *i, 
	       const struct ip_nat_rule_user *find)
{
	/* Compare match and mask */
	return !i->dead
		&& memcmp(&find->match, &i->match, 
			  sizeof(struct ip_conntrack_tuple)) == 0
		&& memcmp(&find->mask, &i->mask, 
			  sizeof(struct ip_conntrack_tuple)) == 0;
}

/* Kernel ip_nat_rule layout:

   [ struct ip_nat_rule ]
   [ rest of range (if > 1 part) ]
   [ moptions (if any) ]
*/
/* Return 0 or -errno */
static int rule_user_to_rule(const struct ip_nat_rule_user *user,
			     struct ip_nat_rule **rule,
			     struct ip_nat_mapping_type *mtype,
			     unsigned int optlen,
			     size_t extra_range)
{
	if (!user->mask.dst.protonum
	    && (user->mask.src.u.all || user->mask.dst.u.all
		|| (user->range.flags & IP_NAT_RANGE_PROTO_SPECIFIED))) {
		DEBUGP("rule_user_to_rule: bad proto specifics\n");
		return -EINVAL;
	}

	*rule = kmalloc(sizeof(struct ip_nat_rule) 
			+ sizeof(struct ip_nat_range)*extra_range
			+ optlen,
#ifdef ALWAYS_IN_INTERRUPT
			GFP_ATOMIC
#else
			GFP_KERNEL
#endif
		);
	if (!*rule)
		return -ENOMEM;

	**rule = ((struct ip_nat_rule){ { NULL, NULL },
					ATOMIC_INIT(1), 0,
					user->match, user->mask,
					user->maniptype,
					precedence(&user->mask),
					mtype, 
					&(*rule)->mr.range[1+extra_range],
					{ 1+extra_range, { user->range } } });
	
	/* Make sure padding is zero. */
	(*rule)->match.src.pad = (*rule)->mask.src.pad = 0;

	/* Copy options (if any) from user */
	memcpy(&(*rule)->mr.range[1+extra_range], user+1, optlen);

	if (mtype && !mtype->check(*rule)) {
		kfree(*rule);
		return -EINVAL;
	}
	return 0;
}

static void insert_rule(struct ip_nat_rule *rule)
{
	IP_CONNTRACK_ASSERT(!rule->dead);
	IP_CONNTRACK_ASSERT(atomic_read(&rule->use) == 1);

	WRITE_LOCK(&ip_nat_lock);
	LIST_INSERT(&nat_rules, rule, prec_cmpfn);
	WRITE_UNLOCK(&ip_nat_lock);
}

static int insert_user_rule(struct ip_nat_rule_user *ruleuser,
			    struct ip_nat_mapping_type *mtype,
			    unsigned int optlen)
{
	struct ip_nat_rule *match, *newrule;
	int ret;

	READ_LOCK(&ip_nat_lock);
	match = LIST_FIND(&nat_rules, rule_usercmpfn, 
			  struct ip_nat_rule *, ruleuser);
	if (match)
		atomic_inc(&match->use);
	READ_UNLOCK(&ip_nat_lock);
	if (match) {
		/* Manipulation type and mapping type must be same;
		 * don't try to merge mappings with options. */
		if (match->maniptype != ruleuser->maniptype
		    || match->mtype != mtype
		    || optlen != 0) {
			DEBUGP("insert_rule: %u vs %u, mtype `%s' vs `%s'\n",
			       match->maniptype, ruleuser->maniptype,
			       match->mtype ? match->mtype->name : "",
			       mtype->name);
			ret = -ENOTEMPTY;
			goto release_match;
		}

		ret = rule_user_to_rule(ruleuser, &newrule, mtype, optlen,
					match->mr.rangesize);
		if (ret == 0) {
			/* Copy old range across. */
			memcpy(&newrule->mr.range[1], match->mr.range, 
			       sizeof(struct ip_nat_range) 
			       * (newrule->mr.rangesize - 1));
			/* Delete home reference. */
			match->dead = 1;
			ip_nat_rule_put(match);

			insert_rule(newrule);
		}

	release_match:
		ip_nat_rule_put(match);
	} else {
		ret = rule_user_to_rule(ruleuser, &newrule, mtype, optlen, 0);
		if (ret == 0) {
			if (mtype && mtype->me)
				__MOD_INC_USE_COUNT(mtype->me);
			insert_rule(newrule);
		}
	}

	return ret;
}

/* This can't be used to delete expected rules, since the precedence
   calculated here won't match the 0 used in ip_nat_expect() above. */
static int delete_rule(const struct ip_nat_rule_user *ruleuser,
		       struct ip_nat_mapping_type *mtype,
		       unsigned int optlen)
{
	struct ip_nat_rule *match, *replacement;
	unsigned int rangenum;
	int ret;

	READ_LOCK(&ip_nat_lock);
	match = LIST_FIND(&nat_rules, rule_usercmpfn, 
			  struct ip_nat_rule *, ruleuser);
	if (match)
		atomic_inc(&match->use);
	READ_UNLOCK(&ip_nat_lock);
	if (!match)
		return -ENOENT;

	/* maniptype and mapping type must be same. */
	if (match->maniptype != ruleuser->maniptype
	    || match->mtype != mtype) {
		DEBUGP("delete_rule: %u vs %u, mtype `%s' vs `%s'\n",
		       match->maniptype, ruleuser->maniptype,
		       match->mtype ? match->mtype->name : "",
		       ruleuser->typename);
		ret = -ENOENT;
		goto out;
	}

	/* Find range in rule. */
	for (rangenum = 0; rangenum < match->mr.rangesize; rangenum++) {
		if (memcmp(&ruleuser->range, &match->mr.range[rangenum],
			   sizeof(ruleuser->range)) == 0) {
			break;
		}
	}
	if (rangenum == match->mr.rangesize) {
		DEBUGP("delete_rule: Found rule, but not matching range\n");
		ret = -ENOENT;
		goto out;
	}

	/* Create and insert replacement.  Locking-wise, easier than
           mangling in place. */
	if (match->mr.rangesize > 1) {
		replacement = kmalloc(sizeof(struct ip_nat_rule) 
				      + (sizeof(struct ip_nat_range)
					 * (match->mr.rangesize-1))
				      + optlen, GFP_KERNEL);
		if (!replacement) {
			DEBUGP("delete_rule: oom\n");
			ret = -ENOMEM;
			goto out;
		}
		
		/* Copy up to deleted range. */
		memcpy(replacement, match, 
		       offsetof(typeof(*match), mr.range[rangenum]));
		/* Copy after deleted range. */
		memcpy(&replacement->mr.range[rangenum],
		       &match->mr.range[rangenum+1],
		       (match->mr.rangesize - rangenum - 1)
		       * sizeof(struct ip_nat_range) + optlen);
		replacement->mr.rangesize = match->mr.rangesize-1;
		atomic_set(&replacement->use, 1);
		replacement->mopts 
			= &replacement->mr.range[replacement->mr.rangesize];

		/* Insert new rule (before deleting old one, to avoid hole) */
		insert_rule(replacement);
	} else {
		/* Deleting it altogether: dec use count. */
		if (mtype && mtype->me)
			__MOD_DEC_USE_COUNT(mtype->me);
	}

	/* Drop home reference. */
	match->dead = 1;
	ip_nat_rule_put(match);

	ret = 0;
 out:
	ip_nat_rule_put(match);
	return ret;
}

static int do_ip_nat_ctl(
#ifdef HAVE_NF_GETSOCKOPT
	struct sock *sk,
#endif
	int cmd, void *user, unsigned int len)
{
	int ret;
	struct ip_nat_rule_user *urule;
	struct ip_nat_mapping_type *mtype;

	/* len > 128000 is a sanity check. */
	if (len < sizeof(struct ip_nat_rule_user) || len > 128000) {
		DEBUGP("IP_NAT_SO_INSERT/DELETE: len %u < %u\n",
		       len, sizeof(struct ip_nat_rule_user));
		return -EINVAL;
	} else if ((urule = kmalloc(len, GFP_KERNEL)) == NULL) {
		DEBUGP("IP_NAT_SO_INSERT/DELETE: oom for len %u\n", len);
		return -ENOMEM;
	} else if (copy_from_user(urule, user, len) != 0) {
		ret = -EFAULT;
		goto free_out;
	}

	/* Must be NUL terminated for sanity */
	urule->typename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN-1] = '\0';

	if (urule->typename[0]) {
		READ_LOCK(&ip_nat_lock);
		mtype = list_named_find(&mapping_types, urule->typename);

		READ_UNLOCK(&ip_nat_lock);
		if (!mtype) {
#ifndef ALWAYS_IN_INTERRUPT
			load_mapping_type(urule->typename);
			READ_LOCK(&ip_nat_lock);
			mtype = list_named_find(&mapping_types, 
						urule->typename);
			READ_UNLOCK(&ip_nat_lock);
#endif
			if (!mtype) {
				DEBUGP("rule_user: no mapping `%s'\n",
				       urule->typename);
				ret = -EINVAL;
				goto free_out;
			}
		}
		if (len-sizeof(struct ip_nat_rule_user) != mtype->optionlen) {
			DEBUGP("IP_NAT_INSERT/DELETE: len %u != %u\n",
			       len - sizeof(struct ip_nat_rule_user),
			       mtype->optionlen);
			ret = -EINVAL;
			goto free_out;
		}
	} else if (len != sizeof(struct ip_nat_rule_user)) {
		DEBUGP("IP_NAT_INSERT/DELETE: normal len %u != %u\n",
		       len, sizeof(struct ip_nat_rule_user));
		ret = -EINVAL;
		goto free_out;
	} else
		mtype = NULL;

	len -= sizeof(struct ip_nat_rule_user);
	if (cmd == IP_NAT_SO_DELETE) {
		ret = delete_rule(urule, mtype, len);
	} else {
		IP_CONNTRACK_ASSERT(cmd == IP_NAT_SO_INSERT);
		ret = insert_user_rule(urule, mtype, len);
	}

 free_out:
	kfree(urule);
	return ret;
}

#ifndef HAVE_NF_GETSOCKOPT
static struct nf_setsockopt_ops ip_nat_sockopts
= { { NULL, NULL }, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX, do_ip_nat_ctl };
#else
static struct nf_sockopt_ops ip_nat_sockopts
= { { NULL, NULL }, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX+1, do_ip_nat_ctl,
    0, 0, NULL };
#endif

#ifdef CONFIG_PROC_FS
static unsigned int getmask(u_int32_t mask)
{
	int i;

	for (i = 0; mask & 0x80000000; i++) 
		mask <<= 1;

	return i;
}

static unsigned int
dump_rule(char *buffer, const struct ip_nat_rule *rule, unsigned int i)
{
	unsigned int len = 0;
	struct ip_nat_protocol *proto;

	if (rule->dead)
		return 0;

	proto = find_nat_proto(rule->match.dst.protonum);

	len += sprintf(buffer + len, "%s [%s] ", 
		       rule->mtype ? rule->mtype->name : "generic",
		       rule->maniptype == IP_NAT_MANIP_SRC ? "SRC" : "DST");
	len += sprintf(buffer + len,
		       "%u.%u.%u.%u/%u->%u.%u.%u.%u/%u ",
		       IP_PARTS(rule->match.src.ip),
		       getmask(ntohl(rule->mask.src.ip)),
		       IP_PARTS(rule->match.dst.ip),
		       getmask(ntohl(rule->mask.dst.ip)));
	if (rule->mask.dst.protonum) {
		len += sprintf(buffer + len, "proto=%u ", 
			       rule->match.dst.protonum);
		len += proto->print(buffer + len, &rule->match, &rule->mask);
	}
	/* Print out range being mapped to. */
	len += sprintf(buffer + len, "TO: ");
	if (rule->mr.range[i].flags & IP_NAT_RANGE_MAP_IPS) {
		len += sprintf(buffer + len, "%u.%u.%u.%u",
			       IP_PARTS(rule->mr.range[i].min_ip));
		if (rule->mr.range[i].max_ip != rule->mr.range[i].min_ip)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS(rule->mr.range[i].max_ip));
		len += sprintf(buffer + len, " ");
	}
	if (rule->mr.range[i].flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		len += proto->print_range(buffer + len, &rule->mr.range[i]);

	len += sprintf(buffer + len, "\n");
	return len;
}

static int 
get_rules(char *buffer, char **start, off_t offset, int length, int dummy)
{
	off_t upto = 0;
	int len = 0,last_len = 0;
	struct list_head *i;

	READ_LOCK(&ip_nat_lock);
	for (i = nat_rules.next; i != &nat_rules; i = i->next) {
		struct ip_nat_rule *rule = (struct ip_nat_rule *)i;
		unsigned int j;

		for (j = 0; j < rule->mr.rangesize; j++, upto++) {
			if (upto < offset) continue;

			last_len = len;
			len += dump_rule(buffer+len, rule, j);
			if (len > length) {
				len = last_len;
				goto out;
			}
		}
	}

 out:
	READ_UNLOCK(&ip_nat_lock);

	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}

struct proc_dir_entry ip_nat_rule_procinfo = {
	0, 12, "ip_nat_rules", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, get_rules
};
#endif

int ip_nat_rule_init(void)
{
	int ret;

	ret = nf_register_sockopt(&ip_nat_sockopts);
	if (ret == 0) {
#ifdef CONFIG_PROC_FS
		proc_net_register(&ip_nat_rule_procinfo);
#endif
		WRITE_LOCK(&ip_nat_lock);
		list_prepend(&mapping_types, &mapping_expect);
		WRITE_UNLOCK(&ip_nat_lock);
	}
	return ret;
}

void ip_nat_rule_cleanup(void)
{
	unsigned long flags;
	nf_unregister_sockopt(&ip_nat_sockopts);

#ifdef CONFIG_PROC_FS
	proc_net_unregister(ip_nat_rule_procinfo.low_ino);
#endif

	/* Delete all the rules. */
	WRITE_LOCK(&ip_nat_lock);
	/* Protect against timers. */
	save_flags(flags);
	cli();
	while (!list_empty(&nat_rules)) {
		struct ip_nat_rule *rule = (void *)nat_rules.next;

		/* Careful: deactivate timers on expect rules! */
		if (rule->mtype == &mapping_expect) {
			struct expect_rule *erule = (void *)rule;
			del_timer(&erule->timer);
		}
		LIST_DELETE(&nat_rules, rule);
		kfree(rule);
	}

	restore_flags(flags);
	LIST_DELETE(&mapping_types, &mapping_expect);
	WRITE_UNLOCK(&ip_nat_lock);
}
