/* This file contains all the functions required for the standalone
   ip_nat module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define EXPORT_SYMTAB
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/spinlock.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_nat_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_nat_lock)

#include "conntrack/struct_reserve.h"
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_rule.h"
#include "NAT/ip_nat_protocol.h"
#include "NAT/ip_nat_core.h"
#include "NAT/ip_nat_helper.h"
#include "NAT/ip_nat_expect.h"
#include "common/listhelp.h"

static unsigned int 
alloc_null_binding(struct ip_conntrack *conntrack,
		   const struct net_device *device)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED). */
	struct ip_nat_multi_range mr = { 1, 
					 { { IP_NAT_RANGE_MAP_IPS, 
					     conntrack->tuplehash[0].tuple.src.ip,
					     conntrack->tuplehash[0].tuple.src.ip,
					     { 0 }, { 0 } } } };

	return ip_nat_setup_info(conntrack, IP_NAT_MANIP_SRC, &mr, device);
}

static unsigned int 
init_conntrack(struct ip_conntrack *conntrack,
	       const struct net_device *device)
{
	struct ip_nat_rule *rule;
	struct ip_nat_info *info = (void *)conntrack + ip_nat_offset;
	unsigned int ret;

	/* Protects the NAT private region of the conntrack. */
	MUST_BE_WRITE_LOCKED(&ip_nat_lock);
	rule = __ip_nat_rule_get(&conntrack->tuplehash[0].tuple);

	/* If there's a rule which matches, set up the translation. */
	if (rule) {
		if (rule->mtype)
			ret = rule->mtype->setup(conntrack, rule->maniptype,
						 &rule->mr, device, 
						 rule->mopts);
		else {
			ret = ip_nat_setup_info(conntrack, rule->maniptype,
						&rule->mr, device);
		}
		ip_nat_rule_put(rule);
	} else {
		/* Set up null or pseudo-null binding. */
		ret = alloc_null_binding(conntrack, device);
	}

	if (ret == NF_ACCEPT) {
		place_in_hashes(conntrack);
		info->initialized = 1;
	}

	return ret;
}

static unsigned int 
ip_nat_fn(unsigned int hooknum,
	  struct sk_buff **pskb,
	  const struct net_device *in,
	  const struct net_device *out)
{
	struct ip_conntrack_tuple_hash *h;
	enum ip_conntrack_info ctinfo;
	struct ip_nat_info *info;
	unsigned int ret = NF_ACCEPT;

	/* FIXME: One day, fill in properly. --RR */
	(*pskb)->nfcache |= NFC_UNKNOWN | NFC_ALTERED;

	/* If we had a hardware checksum before, it's now invalid */
	(*pskb)->ip_summed = CHECKSUM_NONE;

	ctinfo = ip_conntrack_get(*pskb, &h);

	switch (ctinfo) {
	case IP_CT_INVALID:
		DEBUGP("skb %p = IP_CT_INVALID\n", *pskb);
		ret = NF_ACCEPT;
		goto out;
		
	case IP_CT_RELATED:
		if ((*pskb)->nh.iph->protocol == IPPROTO_ICMP) {
			icmp_reply_translation(*pskb, h->ctrack, DIRECTION(h));
			ret = NF_ACCEPT;
			goto release;
		}
		/* Fall thru... */
	case IP_CT_NEW:
		info = (void *)h->ctrack + ip_nat_offset;

		WRITE_LOCK(&ip_nat_lock);
		/* Seen it before?  This can happen for loopback, retrans,
		   or local packets.. */
		if (!info->initialized) {
			/* Device == NULL means a local packet, since
			   it the hook is either NF_IP_PRE_ROUTING or
			   NF_IP_LOCAL_OUT. */
			IP_CONNTRACK_ASSERT(hooknum == NF_IP_PRE_ROUTING
					    || hooknum == NF_IP_LOCAL_OUT);
			ret = init_conntrack(h->ctrack, in);
		}
		WRITE_UNLOCK(&ip_nat_lock);
		if (ret != NF_ACCEPT)
			goto release;
		break;

	default:
		/* ESTABLISHED / ESTABLISHED_SPECIFIES_RELATED */ 
		info = (void *)h->ctrack + ip_nat_offset;
	}

	if (info->dirinfo[DIRECTION(h)].hook == hooknum)
		ret = do_binding(h, ctinfo, pskb);

 release:
	ip_conntrack_put(h->ctrack);
 out:
	return ret;
}

/* We've already marked the conntrack for NAT, but now the packet is
   coming in to the local box, it's simplest to undo it. */
static void ip_nat_undo(struct ip_conntrack *conntrack)
{
	struct ip_conntrack_tuple tuple;

	invert_tuplepr(&tuple, &conntrack->tuplehash[0].tuple);
	ip_conntrack_alter_reply(conntrack, &tuple);
}

static unsigned int 
ip_nat_maybe_not(unsigned int hooknum,
		 struct sk_buff **pskb,
		 const struct net_device *in,
		 const struct net_device *out)
{
	struct ip_conntrack_tuple_hash *h;
	enum ip_conntrack_info ctinfo;

	ctinfo = ip_conntrack_get(*pskb, &h);

	/* Special case: catch source-manip'ed packet coming in */
	if (ctinfo == IP_CT_NEW) {
		struct ip_nat_info *info = (void *)h->ctrack + ip_nat_offset;

		/* Wanted to do source manip, but we don't, because it's
		   coming into the box itself. */
		if (info->dirinfo[0].hook == NF_IP_POST_ROUTING)
			ip_nat_undo(h->ctrack);
	}
	if (ctinfo != IP_CT_INVALID)
		ip_conntrack_put(h->ctrack);

	return NF_ACCEPT;
}

static unsigned int 
ip_nat_out(unsigned int hooknum,
	   struct sk_buff **pskb,
	   const struct net_device *in,
	   const struct net_device *out)
{
	/* We can hit fragment here.  Although conntrack defrags, it
	   also refrags (local packets).  Ignore all but head fragment. */
	if ((*pskb)->nh.iph->frag_off & __constant_htons(IP_OFFSET))
		return NF_ACCEPT;

	return ip_nat_fn(hooknum, pskb, in, out);
}

/* Packet filtering is priority 0: for local_in, we want to be after
   it, for local_out, we want to be before.  And after connection tracking.  */
static struct nf_hook_ops ip_nat_in_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_PRE_ROUTING, 0 };
static struct nf_hook_ops ip_nat_out_ops
= { { NULL, NULL }, ip_nat_out, NULL, PF_INET, NF_IP_POST_ROUTING, -100 };
static struct nf_hook_ops ip_nat_local_in_ops
= { { NULL, NULL }, ip_nat_maybe_not, NULL, PF_INET, NF_IP_LOCAL_IN, 100 };
static struct nf_hook_ops ip_nat_local_out_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_LOCAL_OUT, -100 };

/* Protocol registration. */
int ip_nat_protocol_register(struct ip_nat_protocol *proto)
{
	int ret = 0;
	struct list_head *i;

	WRITE_LOCK(&ip_nat_lock);
	for (i = protos.next; i != &protos; i = i->next) {
		if (((struct ip_nat_protocol *)i)->protonum 
		    == proto->protonum) {
			ret = -EBUSY;
			goto out;
		}
	}

	list_prepend(&protos, proto);
	MOD_INC_USE_COUNT;

 out:
	WRITE_UNLOCK(&ip_nat_lock);
	return ret;
}


/* Noone stores the protocol anywhere; simply delete it. */
void ip_nat_protocol_unregister(struct ip_nat_protocol *proto)
{
	WRITE_LOCK(&ip_nat_lock);
	LIST_DELETE(&protos, proto);
	WRITE_UNLOCK(&ip_nat_lock);

	/* Someone could be still looking at the proto in a bh.  Wait
           for them. */
	synchronize_bh();

	MOD_DEC_USE_COUNT;
}


static int init_or_cleanup(int init)
{
	int ret = 0;

	if (!init) goto cleanup;

	ret = ip_nat_rule_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_nothing;
	}
	ret = ip_nat_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_rule_init;
	}

	ret = nf_register_hook(&ip_nat_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_nat;
	}
	ret = nf_register_hook(&ip_nat_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register out hook.\n");
		goto cleanup_inops;
	}
	ret = nf_register_hook(&ip_nat_local_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local in hook.\n");
		goto cleanup_outops;
	}
	ret = nf_register_hook(&ip_nat_local_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local out hook.\n");
		goto cleanup_localinops;
	}
	__MOD_INC_USE_COUNT(ip_conntrack_module);
	return ret;

 cleanup:
	__MOD_DEC_USE_COUNT(ip_conntrack_module);
	nf_unregister_hook(&ip_nat_local_out_ops);
 cleanup_localinops:
	nf_unregister_hook(&ip_nat_local_in_ops);
 cleanup_outops:
	nf_unregister_hook(&ip_nat_out_ops);
 cleanup_inops:
	nf_unregister_hook(&ip_nat_in_ops);
 cleanup_nat:
	ip_nat_cleanup();
 cleanup_rule_init:
	ip_nat_rule_cleanup();
 cleanup_nothing:
	MUST_BE_READ_WRITE_UNLOCKED(&ip_nat_lock);
	MUST_BE_READ_WRITE_UNLOCKED(&ip_conntrack_lock);
	return ret;
}

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
	init_or_cleanup(0);
}

EXPORT_SYMBOL(ip_nat_setup_info);
EXPORT_SYMBOL(ip_nat_register_mapping_type);
EXPORT_SYMBOL(ip_nat_unregister_mapping_type);
EXPORT_SYMBOL(ip_nat_helper_register);
EXPORT_SYMBOL(ip_nat_helper_unregister);
EXPORT_SYMBOL(ip_nat_expect);
EXPORT_SYMBOL(ip_nat_cheat_check);

